from ._async_call import AlcallParams, BcallParams, alcall, bcall
from ._hash import hash_dict
from ._list_call import LcallParams, lcall
from ._models import DataClass, Params
from ._to_list import ToListParams, to_list
from ._types import (
    Enum,
    KeysDict,
    MaybeSentinel,
    MaybeUndefined,
    MaybeUnset,
    SingletonType,
    T,
    Undefined,
    UndefinedType,
    Unset,
    UnsetType,
    is_sentinel,
    not_sentinel,
)
from .concurrency import *

__all__ = (
    "Undefined",
    "Unset",
    "MaybeUndefined",
    "MaybeUnset",
    "MaybeSentinel",
    "SingletonType",
    "UndefinedType",
    "UnsetType",
    "KeysDict",
    "T",
    "Enum",
    "is_sentinel",
    "not_sentinel",
    "Params",
    "DataClass",
    "Enum",
    "hash_dict",
    "to_list",
    "ToListParams",
    "lcall",
    "LcallParams",
    "alcall",
    "bcall",
    "AlcallParams",
    "BcallParams",
)



from dataclasses import dataclass, field
from typing import Any, ClassVar

from typing_extensions import override

from ._types import Undefined, Unset, is_sentinel

__all__ = ("Params", "DataClass")


@dataclass(slots=True, frozen=True, init=False)
class Params:
    """Base class for parameters used in various functions."""

    _none_as_sentinel: ClassVar[bool] = False
    """If True, None is treated as a sentinel value."""

    _strict: ClassVar[bool] = False
    """No sentinels allowed if strict is True."""

    _prefill_unset: ClassVar[bool] = True
    """If True, unset fields are prefilled with Unset."""

    _allowed_keys: ClassVar[set[str]] = field(
        default=set(), init=False, repr=False
    )
    """Class variable cache to store allowed keys for parameters."""

    def __init__(self, **kwargs: Any):
        """Initialize the Params object with keyword arguments."""
        # Set all attributes from kwargs, allowing for sentinel values
        for k, v in kwargs.items():
            if k in self.allowed():
                object.__setattr__(self, k, v)
            else:
                raise ValueError(f"Invalid parameter: {k}")

        # Validate after setting all attributes
        self._validate()

    @classmethod
    def _is_sentinel(cls, value: Any) -> bool:
        """Check if a value is a sentinel (Undefined or Unset)."""
        if value is None and cls._none_as_sentinel:
            return True
        return is_sentinel(value)

    @classmethod
    def allowed(cls) -> set[str]:
        """Return the keys of the parameters."""
        if cls._allowed_keys:
            return cls._allowed_keys
        cls._allowed_keys = {
            i for i in cls.__dataclass_fields__.keys() if not i.startswith("_")
        }
        return cls._allowed_keys

    @override
    def _validate(self) -> None:
        def _validate_strict(k):
            if self._strict and self._is_sentinel(getattr(self, k, Unset)):
                raise ValueError(f"Missing required parameter: {k}")
            if (
                self._prefill_unset
                and getattr(self, k, Undefined) is Undefined
            ):
                object.__setattr__(self, k, Unset)

        for k in self.allowed():
            _validate_strict(k)

    def default_kw(self) -> Any:

        # create a partial function with the current parameters
        dict_ = self.to_dict()

        # handle kwargs if present, handle both 'kwargs' and 'kw'
        kw_ = {}
        kw_.update(dict_.pop("kwargs", {}))
        kw_.update(dict_.pop("kw", {}))
        dict_.update(kw_)
        return dict_

    def to_dict(self) -> dict[str, str]:
        data = {}
        for k in self.allowed():
            if not self._is_sentinel(v := getattr(self, k, Undefined)):
                data[k] = v
        return data


@dataclass(slots=True)
class DataClass:
    """A base class for data classes with strict parameter handling."""

    _none_as_sentinel: ClassVar[bool] = False
    """If True, None is treated as a sentinel value."""

    _strict: ClassVar[bool] = False
    """No sentinels allowed if strict is True."""

    _prefill_unset: ClassVar[bool] = True
    """If True, unset fields are prefilled with Unset."""

    _allowed_keys: ClassVar[set[str]] = field(
        default=set(), init=False, repr=False
    )
    """Class variable cache to store allowed keys for parameters."""

    def __post_init__(self):
        """Post-initialization to ensure all fields are set."""
        self._validate()

    @classmethod
    def allowed(cls) -> set[str]:
        """Return the keys of the parameters."""
        if cls._allowed_keys:
            return cls._allowed_keys
        cls._allowed_keys = {
            i for i in cls.__dataclass_fields__.keys() if not i.startswith("_")
        }
        return cls._allowed_keys

    @override
    def _validate(self) -> None:
        def _validate_strict(k):
            if self._strict and self._is_sentinel(getattr(self, k, Unset)):
                raise ValueError(f"Missing required parameter: {k}")
            if (
                self._prefill_unset
                and getattr(self, k, Undefined) is Undefined
            ):
                self.__setattr__(k, Unset)

        for k in self.allowed():
            _validate_strict(k)

    def to_dict(self) -> dict[str, str]:
        data = {}
        print(self.allowed())
        for k in type(self).allowed():
            if not self._is_sentinel(v := getattr(self, k)):
                data[k] = v
        return data

    @classmethod
    def _is_sentinel(cls, value: Any) -> bool:
        """Check if a value is a sentinel (Undefined or Unset)."""
        if value is None and cls._none_as_sentinel:
            return True
        return is_sentinel(value)



"""Structured concurrency primitives for pynector.

This module provides structured concurrency primitives using AnyIO,
which allows for consistent behavior across asyncio and trio backends.
"""

from .cancel import CancelScope, fail_after, move_on_after
from .errors import get_cancelled_exc_class, shield
from .patterns import (
    ConnectionPool,
    WorkerPool,
    parallel_requests,
    retry_with_timeout,
)
from .primitives import CapacityLimiter, Condition, Event, Lock, Semaphore
from .resource_tracker import (
    ResourceTracker,
    cleanup_check,
    get_global_tracker,
    resource_leak_detector,
    track_resource,
    untrack_resource,
)
from .task import TaskGroup, create_task_group
from .utils import is_coro_func

ConcurrencyEvent = Event
__all__ = (
    "TaskGroup",
    "create_task_group",
    "CancelScope",
    "move_on_after",
    "fail_after",
    "ConnectionPool",
    "WorkerPool",
    "parallel_requests",
    "retry_with_timeout",
    "Lock",
    "Semaphore",
    "CapacityLimiter",
    "Event",
    "Condition",
    "get_cancelled_exc_class",
    "shield",
    "ResourceTracker",
    "resource_leak_detector",
    "track_resource",
    "untrack_resource",
    "cleanup_check",
    "get_global_tracker",
    "is_coro_func",
    "ConcurrencyEvent",
)



"""Resource management primitives for structured concurrency.

Pure async primitives focused on correctness and simplicity.
"""

import math
from types import TracebackType

import anyio

from .resource_tracker import track_resource, untrack_resource


class Lock:
    """A mutex lock for controlling access to a shared resource."""

    def __init__(self):
        """Initialize a new lock."""
        self._lock = anyio.Lock()
        self._acquired = False
        track_resource(self, f"Lock-{id(self)}", "Lock")

    def __del__(self):
        """Clean up resource tracking when lock is destroyed."""
        try:
            untrack_resource(self)
        except Exception:
            pass

    async def __aenter__(self) -> None:
        """Acquire the lock."""
        await self._lock.acquire()
        self._acquired = True

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Release the lock."""
        self._lock.release()
        self._acquired = False

    async def acquire(self) -> None:
        """Acquire the lock directly."""
        await self._lock.acquire()
        self._acquired = True

    def release(self) -> None:
        """Release the lock directly."""
        if not self._acquired:
            raise RuntimeError(
                "Attempted to release lock that was not acquired by this task"
            )
        self._lock.release()
        self._acquired = False


class Semaphore:
    """A semaphore preventing excessive releases."""

    def __init__(self, initial_value: int):
        """Initialize a new semaphore."""
        if initial_value < 0:
            raise ValueError("The initial value must be >= 0")
        self._initial_value = initial_value
        self._current_acquisitions = 0
        self._semaphore = anyio.Semaphore(initial_value)
        track_resource(self, f"Semaphore-{id(self)}", "Semaphore")

    def __del__(self):
        """Clean up resource tracking when semaphore is destroyed."""
        try:
            untrack_resource(self)
        except Exception:
            pass

    async def __aenter__(self) -> None:
        """Acquire the semaphore."""
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Release the semaphore."""
        self.release()

    async def acquire(self) -> None:
        """Acquire the semaphore."""
        await self._semaphore.acquire()
        self._current_acquisitions += 1

    def release(self) -> None:
        """Release the semaphore."""
        if self._current_acquisitions <= 0:
            raise RuntimeError(
                "Cannot release semaphore: no outstanding acquisitions"
            )
        self._semaphore.release()
        self._current_acquisitions -= 1

    @property
    def current_acquisitions(self) -> int:
        """Get the current number of outstanding acquisitions."""
        return self._current_acquisitions

    @property
    def initial_value(self) -> int:
        """Get the initial semaphore value."""
        return self._initial_value


class CapacityLimiter:
    """A context manager for limiting the number of concurrent operations."""

    def __init__(self, total_tokens: int | float):
        """Initialize a new capacity limiter."""
        if total_tokens == math.inf:
            processed_tokens = math.inf
        elif isinstance(total_tokens, (int, float)) and total_tokens >= 1:
            processed_tokens = (
                int(total_tokens) if total_tokens != math.inf else math.inf
            )
        else:
            raise ValueError(
                "The total number of tokens must be >= 1 (int or math.inf)"
            )

        self._limiter = anyio.CapacityLimiter(processed_tokens)
        self._borrower_counter = 0
        self._active_borrowers = {}
        track_resource(self, f"CapacityLimiter-{id(self)}", "CapacityLimiter")

    def __del__(self):
        """Clean up resource tracking when limiter is destroyed."""
        try:
            untrack_resource(self)
        except Exception:
            pass

    async def __aenter__(self) -> None:
        """Acquire a token."""
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Release the token."""
        self.release()

    async def acquire(self) -> None:
        """Acquire a token."""
        # Create a unique borrower identity for each acquisition
        self._borrower_counter += 1
        borrower = f"borrower-{self._borrower_counter}"
        await self._limiter.acquire_on_behalf_of(borrower)
        self._active_borrowers[borrower] = True

    def release(self) -> None:
        """Release a token."""
        # Find and release the first active borrower
        if not self._active_borrowers:
            raise RuntimeError("No tokens to release")

        borrower = next(iter(self._active_borrowers))
        self._limiter.release_on_behalf_of(borrower)
        del self._active_borrowers[borrower]

    @property
    def total_tokens(self) -> int | float:
        """The total number of tokens."""
        return float(self._limiter.total_tokens)

    @total_tokens.setter
    def total_tokens(self, value: int | float) -> None:
        """Set the total number of tokens."""
        if value == math.inf:
            processed_value = math.inf
        elif isinstance(value, (int, float)) and value >= 1:
            processed_value = int(value) if value != math.inf else math.inf
        else:
            raise ValueError(
                "The total number of tokens must be >= 1 (int or math.inf)"
            )

        current_borrowed = self._limiter.borrowed_tokens
        if processed_value != math.inf and processed_value < current_borrowed:
            raise ValueError(
                f"Cannot set total_tokens to {processed_value}: {current_borrowed} tokens "
                f"are currently borrowed. Wait for tokens to be released or "
                f"set total_tokens to at least {current_borrowed}."
            )

        self._limiter.total_tokens = processed_value

    @property
    def borrowed_tokens(self) -> int:
        """The number of tokens currently borrowed."""
        return self._limiter.borrowed_tokens

    @property
    def available_tokens(self) -> int | float:
        """The number of tokens currently available."""
        return self._limiter.available_tokens


class Event:
    """An event object for task synchronization."""

    def __init__(self):
        """Initialize a new event in the unset state."""
        self._event = anyio.Event()
        track_resource(self, f"Event-{id(self)}", "Event")

    def __del__(self):
        """Clean up resource tracking when event is destroyed."""
        try:
            untrack_resource(self)
        except Exception:
            pass

    def is_set(self) -> bool:
        """Check if the event is set."""
        return self._event.is_set()

    def set(self) -> None:
        """Set the event, allowing all waiting tasks to proceed."""
        self._event.set()

    async def wait(self) -> None:
        """Wait until the event is set."""
        await self._event.wait()


class Condition:
    """A condition variable for task synchronization."""

    def __init__(self, lock: Lock | None = None):
        """Initialize a new condition."""
        self._lock = lock or Lock()
        self._condition = anyio.Condition(self._lock._lock)
        track_resource(self, f"Condition-{id(self)}", "Condition")

    def __del__(self):
        """Clean up resource tracking when condition is destroyed."""
        try:
            untrack_resource(self)
        except Exception:
            pass

    async def __aenter__(self) -> "Condition":
        """Acquire the underlying lock."""
        await self._lock.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Release the underlying lock."""
        await self._lock.__aexit__(exc_type, exc_val, exc_tb)

    async def wait(self) -> None:
        """Wait for a notification.

        This releases the underlying lock, waits for a notification, and then
        reacquires the lock.
        """
        await self._condition.wait()

    async def notify(self, n: int = 1) -> None:
        """Notify waiting tasks.

        Args:
            n: The number of tasks to notify
        """
        await self._condition.notify(n)

    async def notify_all(self) -> None:
        """Notify all waiting tasks."""
        await self._condition.notify_all()



from __future__ import annotations

from enum import Enum as _Enum
from typing import Any, Final, Literal, TypeVar, Union

from typing_extensions import TypedDict

__all__ = (
    "Undefined",
    "Unset",
    "MaybeUndefined",
    "MaybeUnset",
    "MaybeSentinel",
    "SingletonType",
    "UndefinedType",
    "UnsetType",
    "KeysDict",
    "T",
    "Enum",
    "is_sentinel",
    "not_sentinel",
)

T = TypeVar("T")


class _SingletonMeta(type):
    """Metaclass that guarantees exactly one instance per subclass.

    This ensures that sentinel values maintain identity across the entire application,
    allowing safe identity checks with 'is' operator.
    """

    _cache: dict[type, SingletonType] = {}

    def __call__(cls, *a, **kw):
        if cls not in cls._cache:
            cls._cache[cls] = super().__call__(*a, **kw)
        return cls._cache[cls]


class SingletonType(metaclass=_SingletonMeta):
    """Base class for singleton sentinel types.

    Provides consistent interface for sentinel values with:
    - Identity preservation across deepcopy
    - Falsy boolean evaluation
    - Clear string representation
    """

    __slots__: tuple[str, ...] = ()

    def __deepcopy__(self, memo):  # copy & deepcopy both noop
        return self

    def __copy__(self):
        return self

    # concrete classes *must* override the two methods below
    def __bool__(self) -> bool: ...
    def __repr__(self) -> str: ...


class UndefinedType(SingletonType):
    """Sentinel for a key or field entirely missing from a namespace.

    Use this when:
    - A field has never been set
    - A key doesn't exist in a mapping
    - A value is conceptually undefined (not just unset)

    Example:
        >>> d = {"a": 1}
        >>> d.get("b", Undefined) is Undefined
        True
    """

    __slots__ = ()

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> Literal["Undefined"]:
        return "Undefined"

    def __str__(self) -> Literal["Undefined"]:
        return "Undefined"


class UnsetType(SingletonType):
    """Sentinel for a key present but value not yet provided.

    Use this when:
    - A parameter exists but hasn't been given a value
    - Distinguishing between None and "not provided"
    - API parameters that are optional but need explicit handling

    Example:
        >>> def func(param=Unset):
        ...     if param is not Unset:
        ...         # param was explicitly provided
        ...         process(param)
    """

    __slots__ = ()

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> Literal["Unset"]:
        return "Unset"

    def __str__(self) -> Literal["Unset"]:
        return "Unset"


Undefined: Final = UndefinedType()
"""A key or field entirely missing from a namespace"""
Unset: Final = UnsetType()
"""A key present but value not yet provided."""

MaybeUndefined = Union[T, UndefinedType]
MaybeUnset = Union[T, UnsetType]
MaybeSentinel = Union[T, UndefinedType, UnsetType]


def is_sentinel(value: Any) -> bool:
    """Check if a value is any sentinel (Undefined or Unset)."""
    return value is Undefined or value is Unset


def not_sentinel(value: Any) -> bool:
    """Check if a value is NOT a sentinel. Useful for filtering operations."""
    return value is not Undefined and value is not Unset


class Enum(_Enum):
    @classmethod
    def allowed(cls) -> tuple[str, ...]:
        return tuple(e.value for e in cls)


class KeysDict(TypedDict, total=False):
    """TypedDict for keys dictionary."""

    key: Any  # Represents any key-type pair



"""Resource tracking utilities for concurrency primitives.

This module provides lightweight resource leak detection and lifecycle tracking
to address the security vulnerabilities identified in the hardening tests.
"""

import weakref
from dataclasses import dataclass
from datetime import datetime
from typing import Any


@dataclass
class ResourceInfo:
    """Information about a tracked resource."""

    name: str
    creation_time: datetime
    resource_type: str


class ResourceTracker:
    """Lightweight resource lifecycle tracking for leak detection.

    This addresses the over-engineering concerns by providing simple,
    practical resource management without complex abstraction layers.
    """

    def __init__(self):
        """Initialize a new resource tracker."""
        self._active_resources: dict[int, ResourceInfo] = {}
        self._weak_refs: weakref.WeakKeyDictionary = (
            weakref.WeakKeyDictionary()
        )

    def track(
        self, resource: Any, name: str, resource_type: str | None = None
    ) -> None:
        """Track a resource for leak detection.

        Args:
            resource: The resource to track
            name: Human-readable name for the resource
            resource_type: Optional type classification
        """
        if resource_type is None:
            resource_type = type(resource).__name__

        resource_info = ResourceInfo(
            name=name,
            creation_time=datetime.now(),
            resource_type=resource_type,
        )

        # Use weak reference to avoid interfering with garbage collection
        self._weak_refs[resource] = resource_info
        self._active_resources[id(resource)] = resource_info

    def untrack(self, resource: Any) -> None:
        """Manually untrack a resource.

        Args:
            resource: The resource to stop tracking
        """
        resource_id = id(resource)
        self._active_resources.pop(resource_id, None)
        self._weak_refs.pop(resource, None)

    def cleanup_check(self) -> list[ResourceInfo]:
        """Check for potentially leaked resources.

        Returns:
            List of resource info for resources that may have leaked
        """
        # Clean up references to garbage collected objects
        current_resources = []
        for resource, info in list(self._weak_refs.items()):
            current_resources.append(info)

        return current_resources

    def get_active_count(self) -> int:
        """Get the number of currently tracked resources.

        Returns:
            Number of active tracked resources
        """
        return len(self._weak_refs)

    def get_resource_summary(self) -> dict[str, int]:
        """Get a summary of tracked resources by type.

        Returns:
            Dictionary mapping resource types to counts
        """
        summary = {}
        for info in self._weak_refs.values():
            resource_type = info.resource_type
            summary[resource_type] = summary.get(resource_type, 0) + 1
        return summary


# Global tracker instance for convenience
_global_tracker = ResourceTracker()


def track_resource(
    resource: Any, name: str, resource_type: str | None = None
) -> None:
    """Track a resource using the global tracker.

    Args:
        resource: The resource to track
        name: Human-readable name for the resource
        resource_type: Optional type classification
    """
    _global_tracker.track(resource, name, resource_type)


def untrack_resource(resource: Any) -> None:
    """Untrack a resource using the global tracker.

    Args:
        resource: The resource to stop tracking
    """
    _global_tracker.untrack(resource)


def get_global_tracker() -> ResourceTracker:
    """Get the global resource tracker instance.

    Returns:
        The global ResourceTracker instance
    """
    return _global_tracker


def cleanup_check() -> list[ResourceInfo]:
    """Check for potentially leaked resources using global tracker.

    Returns:
        List of resource info for resources that may have leaked
    """
    return _global_tracker.cleanup_check()


class resource_leak_detector:
    """Context manager for resource leak detection in tests and production.

    Example:
        async with resource_leak_detector() as tracker:
            lock = Lock()
            tracker.track(lock, "test_lock")
            # ... use lock
        # Automatically checks for leaks on exit
    """

    def __init__(self, raise_on_leak: bool = False):
        """Initialize the leak detector.

        Args:
            raise_on_leak: Whether to raise an exception if leaks are detected
        """
        self.raise_on_leak = raise_on_leak
        self.tracker = ResourceTracker()
        self._initial_count = 0

    async def __aenter__(self) -> ResourceTracker:
        """Enter the context and return the tracker."""
        self._initial_count = self.tracker.get_active_count()
        return self.tracker

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the context and check for leaks."""
        leaked_resources = self.tracker.cleanup_check()

        if leaked_resources and self.raise_on_leak:
            resource_summary = self.tracker.get_resource_summary()
            raise RuntimeError(
                f"Resource leak detected: {len(leaked_resources)} resources "
                f"still active. Summary: {resource_summary}"
            )



"""Common concurrency patterns for structured concurrency."""

from __future__ import annotations

import logging
from collections.abc import Awaitable, Callable
from types import TracebackType
from typing import Any, TypeVar

import anyio

from .cancel import move_on_after
from .primitives import CapacityLimiter, Lock
from .resource_tracker import track_resource, untrack_resource
from .task import create_task_group

logger = logging.getLogger(__name__)

T = TypeVar("T")
R = TypeVar("R")
Response = TypeVar("Response")


class ConnectionPool:
    """A pool of reusable connections."""

    def __init__(
        self,
        max_connections: int,
        connection_factory: Callable[[], Awaitable[T]],
    ):
        """Initialize a new connection pool."""
        if max_connections < 1:
            raise ValueError("max_connections must be >= 1")
        if not callable(connection_factory):
            raise ValueError("connection_factory must be callable")

        self._connection_factory = connection_factory
        self._limiter = CapacityLimiter(max_connections)
        self._connections: list[T] = []
        self._lock = Lock()

        track_resource(self, f"ConnectionPool-{id(self)}", "ConnectionPool")

    def __del__(self):
        """Clean up resource tracking."""
        try:
            untrack_resource(self)
        except Exception:
            pass

    async def acquire(self) -> T:
        """Acquire a connection from the pool."""
        await self._limiter.acquire()

        try:
            async with self._lock:
                if self._connections:
                    return self._connections.pop()

            # No pooled connection available, create new one
            return await self._connection_factory()
        except Exception:
            self._limiter.release()
            raise

    async def release(self, connection: T) -> None:
        """Release a connection back to the pool."""
        try:
            async with self._lock:
                self._connections.append(connection)
        finally:
            self._limiter.release()

    async def __aenter__(self) -> ConnectionPool[T]:
        """Enter the connection pool context."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the connection pool context."""
        # Clean up any remaining connections
        async with self._lock:
            self._connections.clear()


async def parallel_requests(
    inputs: list[str],
    func: Callable[[str], Awaitable[Response]],
    max_concurrency: int = 10,
) -> list[Response]:
    """Execute requests in parallel with controlled concurrency.

    Args:
        inputs: List of inputs
        fetch_func: Async function
        max_concurrency: Maximum number of concurrent requests

    Returns:
        List of responses in the same order as inputs
    """
    if not inputs:
        return []

    results: list[Response | None] = [None] * len(inputs)

    async def bounded_fetch(
        semaphore: anyio.Semaphore, idx: int, url: str
    ) -> None:
        async with semaphore:
            results[idx] = await func(url)

    try:
        async with create_task_group() as tg:
            semaphore = anyio.Semaphore(max_concurrency)

            for i, inp in enumerate(inputs):
                await tg.start_soon(bounded_fetch, semaphore, i, inp)
    except BaseException as e:
        # Re-raise the first exception directly instead of ExceptionGroup
        if hasattr(e, "exceptions") and e.exceptions:
            raise e.exceptions[0]
        else:
            raise

    return results  # type: ignore


async def retry_with_timeout(
    func: Callable[[], Awaitable[T]],
    max_retries: int = 3,
    timeout: float = 30.0,
    backoff_factor: float = 1.0,
) -> T:
    """Retry an async function with exponential backoff and timeout.

    Args:
        func: The async function to retry
        max_retries: Maximum number of retries
        timeout: Timeout for each attempt
        backoff_factor: Multiplier for exponential backoff

    Returns:
        The result of the successful function call

    Raises:
        Exception: The last exception raised by the function
    """
    last_exception = None

    for attempt in range(max_retries):
        try:
            with move_on_after(timeout) as cancel_scope:
                result = await func()
                if not cancel_scope.cancelled_caught:
                    return result
                else:
                    raise TimeoutError(f"Function timed out after {timeout}s")
        except Exception as e:
            last_exception = e
            if attempt < max_retries - 1:
                delay = backoff_factor * (2**attempt)
                await anyio.sleep(delay)
            continue

    if last_exception:
        raise last_exception
    else:
        raise RuntimeError("Retry failed without capturing exception")


class WorkerPool:
    """A pool of worker tasks that process items from a queue."""

    def __init__(
        self, num_workers: int, worker_func: Callable[[Any], Awaitable[None]]
    ):
        """Initialize a new worker pool."""
        if num_workers < 1:
            raise ValueError("num_workers must be >= 1")
        if not callable(worker_func):
            raise ValueError("worker_func must be callable")

        self._num_workers = num_workers
        self._worker_func = worker_func
        self._queue = anyio.create_memory_object_stream(1000)
        self._task_group = None

        track_resource(self, f"WorkerPool-{id(self)}", "WorkerPool")

    def __del__(self):
        """Clean up resource tracking."""
        try:
            untrack_resource(self)
        except Exception:
            pass

    async def start(self) -> None:
        """Start the worker pool."""
        if self._task_group is not None:
            raise RuntimeError("Worker pool is already started")

        self._task_group = create_task_group()
        await self._task_group.__aenter__()

        # Start worker tasks
        for i in range(self._num_workers):
            await self._task_group.start_soon(self._worker_loop)

    async def stop(self) -> None:
        """Stop the worker pool."""
        if self._task_group is None:
            return

        # Close the queue to signal workers to stop
        await self._queue[0].aclose()

        # Wait for all workers to finish
        try:
            await self._task_group.__aexit__(None, None, None)
        finally:
            self._task_group = None

    async def submit(self, item: Any) -> None:
        """Submit an item for processing."""
        if self._task_group is None:
            raise RuntimeError("Worker pool is not started")
        await self._queue[0].send(item)

    async def _worker_loop(self) -> None:
        """Main loop for worker tasks."""
        try:
            async with self._queue[1]:
                async for item in self._queue[1]:
                    try:
                        await self._worker_func(item)
                    except Exception as e:
                        logger.error(f"Worker error processing item: {e}")
        except anyio.ClosedResourceError:
            # Queue was closed, worker should exit gracefully
            pass

    async def __aenter__(self) -> WorkerPool:
        """Enter the worker pool context."""
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the worker pool context."""
        await self.stop()



from lionagi.utils import check_import, is_import_installed

_HAS_PDF2IMAGE = is_import_installed("pdf2image")


def pdf_to_images(
    pdf_path: str, output_folder: str, dpi: int = 300, fmt: str = "jpeg"
) -> list:
    """
    Convert a PDF file into images, one image per page.

    Args:
        pdf_path (str): Path to the input PDF file.
        output_folder (str): Directory to save the output images.
        dpi (int): Dots per inch (resolution) for conversion (default: 300).
        fmt (str): Image format (default: 'jpeg'). Use 'png' if preferred.

    Returns:
        list: A list of file paths for the saved images.
    """
    if not _HAS_PDF2IMAGE:
        raise ModuleNotFoundError(
            "pdf2image is not installed, please install it with `pip install lionagi[unstructured]`"
        )

    import os

    convert_from_path = check_import(
        "pdf2image", import_name="convert_from_path"
    )

    # Ensure the output folder exists
    os.makedirs(output_folder, exist_ok=True)

    # Convert PDF to a list of PIL Image objects
    images = convert_from_path(pdf_path, dpi=dpi)

    saved_paths = []
    for i, image in enumerate(images):
        # Construct the output file name
        image_file = os.path.join(output_folder, f"page_{i + 1}.{fmt}")
        image.save(image_file, fmt.upper())
        saved_paths.append(image_file)

    return saved_paths



def check_async_postgres_available():
    try:
        import sqlalchemy as sa
        from pydapter.extras.async_postgres_ import AsyncPostgresAdapter
        from sqlalchemy.ext.asyncio import create_async_engine

        return True
    except Exception:
        return ImportError(
            "This adapter requires postgres option to be installed. "
            'Please install them using `uv pip install "lionagi[postgres]"`.'
        )


def check_postgres_available():
    try:
        from pydapter.model_adapters.postgres_model import PostgresModelAdapter
        from sqlalchemy import String
        from sqlalchemy.orm import DeclarativeBase

        return True
    except Exception:
        return ImportError(
            "This adapter requires postgres option to be installed. "
            'Please install them using `uv pip install "lionagi[postgres]"`.'
        )
